%%%============================================================================
%%% File     : burm.hrl
%%% Purpose  : BURM support procedures.
%%% Author   : j@svirfneblin.org
%%% License  : BSD
%%% Created  : 2007-08-30 09:58:20 j
%%% Modified : 2007-11-01 18:31:35 j
%%%============================================================================

%%%_. Attributes

-include("eburg_shared.hrl").

%% For conditional tracing
-ifdef(DEBUG).
-define(dbg(Fmt, Data), io:format(Fmt, Data)).
-else.
-define(dbg(Fmt, Data), true).
-endif.


%%%_. State labels

%% Labels are stored with the input tree, in the field accessed by the user
%% provided `state_label' and `state_label_set' functions.
-record(burm_state, {cost, % tuple
                     rule, % tuple
                     id}). % int


%% ?burm_cost(Nt: int, Node: node) -> int
%% returns the cost of deriving `Nt' at `Node'.
-define(burm_cost(Nt, Node),
        element(Nt, (state_label(Node))#burm_state.cost)).


%% ?burm_rule(Nt: int, Node: node) -> int
%% returns the optimal rule for deriving `Nt' at `Node'.
-define(burm_rule(Nt, Node),
        burm_rule(Nt, (state_label(Node))#burm_state.rule)).


%%%_. Labeler
%%% `label' uses `burm_state' and friends to label an entire subject tree
%%% bottom-up. Some notes:
%%%   * It uses three ETS tables for memoization.
%%%   * State IDs are like to BURG's state numbers, only generated at run-time.
%%%   * Our SWDCs are quite different from the SWDCs in the ``On-Demand
%%%     Tree-Parsing Automata'' paper.
%%%   * We spawn a labeler process to isolate the ETS stuff.

%% label(Tree: node) -> node
%% spawns a labeler process if necessary, uses this process to
%% label `Tree'.
label(Tree) ->
    case whereis(burm_labeler) of
        undefined ->
            Self = self(),
            Pid  = spawn_opt(fun() -> burm_labeler(Self) end,
                             [{min_heap_size, 10000}]),
            burm_rpc(Pid, Tree);
        Pid ->
            burm_rpc(Pid, Tree)
    end.


%% burm_labeler(Parent: pid)
%% sets up ETS tables then enters an infinite loop.
burm_labeler(Parent) ->
    register(burm_labeler, self()),
    case burm_maybe_load() of
        true ->
            ?dbg("ETS tables loaded~n", []);
        false ->
            ?dbg("creating ETS tables~n", []),
            %% #burm_state -> int
            ets:new(burm_id_cache, [named_table]),
            %% {atom, int, ..., int} -> #burm_state
            ets:new(burm_state_cache, [named_table]),
            %% {{atom, int, ..., int}, {int, ..., int}} -> #burm_state
            ets:new(burm_swdc_cache, [named_table])
    end,
    burm_loop(Parent).


%% burm_rpc(Pid: pid, Tree: node) -> node
%% labels `Tree' via `burm_labeler'.
burm_rpc(Pid, Tree) ->
    Pid ! Tree,
    receive
        Err={burm_exn, _Reason} -> throw(Err);
        Result -> Result
    end.


%% burm_loop(Pid: pid)
%% waits until somebody sends it a tree, which it labels and sends to `Pid'.
burm_loop(Pid) ->
    receive
        Tree ->
            case catch burm_label(Tree) of
                Err={burm_exn, _Reason} -> Pid ! Err;
                Result -> Pid ! Result
            end
    end,
    burm_loop(Pid).


%% label(Tree: node) -> node
%% returns `Tree' with the `state_label' field set.
burm_label(Tree0) ->
    Op = op_label(Tree0),

    ?dbg("~p: labeling kids~n", [Op]),
    Kids0  = kids(Tree0),
    Kids1  = lists:map(fun burm_label/1, Kids0),
    Tree1  = kids_set(Tree0, Kids1),
    ?dbg("~p: kids labeled~n", [Op]),

    Key0 = list_to_tuple([Op] ++
                         [(state_label(K))#burm_state.id || K <- Kids1]),

    Label =
        case burm_is_tainted(Op) of
            true ->
                Key1 = {Key0, burm_dyncost(Op, Tree1)},
                ?dbg("~p: Key1 is ~p~n", [Op, Key1]),
                case ets:lookup(burm_swdc_cache, Key1) of
                    [] ->
                        L = burm_calc_label(Tree1),
                        ets:insert(burm_swdc_cache, {Key1, L}),
                        ?dbg("~p: swdc calculated and saved: ~p~n~n", [Op, L]),
                        L;
                    [{Key1, L}] ->
                        ?dbg("~p: swdc found: ~p~n~n", [Op, L]),
                        L;
                    _ ->
                        burm_panic({lookup, swdc_cache})
                end;
            false ->
                case ets:lookup(burm_state_cache, Key0) of
                    [] ->
                        L = burm_calc_label(Tree1),
                        ets:insert(burm_state_cache, {Key0, L}),
                        ?dbg("~p: state calculated and saved: ~p~n~n", [Op, L]),
                        L;
                    [{Key0, L}] ->
                        ?dbg("~p: state found: ~p~n~n", [Op, L]),
                        L;
                    _ ->
                        burm_panic({lookup, state_cache})
                end
        end,
    state_label_set(Tree1, Label).



%% burm_calc_label(Node: node) -> #burm_state
%% returns the full state label for `Node'.
burm_calc_label(Node) ->
    Label = burm_normalize(burm_state(Node)),
    Id    = burm_maybe_reuse_id(Label),
    Label#burm_state{id=Id}.


%% burm_normalize(S: #burm_state) -> #burm_state
%% returns `S' with its cost field normalized (we only record the costs of
%% nonterminals relative to each other, not the total costs; this allows us
%% to reuse ids more frequently).
burm_normalize(S=#burm_state{cost=Cost}) ->
    C0    = tuple_to_list(Cost),
    Delta = lists:min(C0),
    C1    = lists:map(
              fun(C) ->
                      if C =:= ?INFINITY -> C;
                         true -> C - Delta
                      end
              end,
              C0),
    S#burm_state{cost=list_to_tuple(C1)}.


%% burm_maybe_reuse_id(Label: {tulpe, tuple}) -> int
%% returns the state number for `Label'.
burm_maybe_reuse_id(Label) ->
    case ets:lookup(burm_id_cache, Label) of
        [] ->
            Id = burm_mk_id(),
            ets:insert(burm_id_cache, {Label, Id}),
            ?dbg("new id: ~p~n", [Id]),
            Id;
        [{Label, Id}] ->
            ?dbg("reusing id: ~p~n", [Id]),
            Id;
        _ -> burm_panic({lookup, id_cache})
    end.


%% burm_mk_id() -> int
%% returns a new state id.
burm_mk_id() ->
    N = case get(burm_id) of
            undefined -> 0;
            X -> X+1
        end,
    put(burm_id, N),
    N.


%%%_. Reducer
%%% `reduce' reduces a labeled tree top-down. Actions are fired in post order.
%%% It returns whatever the semantic action of the rule used to reduce the
%%% input node returns. Its second argument is the INN of some nonterminal;
%%% when using `reduce' directly (rather than calling `rewrite' which calls
%%% `reduce' with the INN of the start nonterminal), the INN for nonterminal NT
%%% may be obtained by writing ?NT (eburg generates macro definitions mapping
%%% nonterminals to their INNs).

%% reduce(Tree: node, Nt: int) -> term
%% returns the result of executing the user-supplied semantic actions
%% associated with `Tree'.
reduce(Tree, Nt) ->
    Op    = op_label(Tree),
    Rule  = ?burm_rule(Nt, Tree),

    if Rule =:= ?INFINITY -> burm_panic(nomatch);
       true               -> ok
    end,

    Kids0 = kids(Tree),
    Nts   = burm_nts(Rule),
    ?dbg("~p: reducing to ~s~n", [Op, burm_string({nt, Nt})]),
    ?dbg("~p: rule is ~s~n", [Op, burm_string({rule, Rule})]),
    case burm_is_chainrule(Rule) of
        true ->
            ?dbg("~p: chainrule, recursing~n", [Op]),
            KX = reduce(Tree, hd(Nts)), % Nts is singleton list
            ?dbg("~p: firing action~n~n", [Op]),
            burm_action(Tree, [KX], Rule);
        false ->
            ?dbg("~p: baserule, reducing kids~n", [Op]),
            Kids1 = [reduce(T, N) || {T, N} <- lists:zip(Kids0, Nts)],
            ?dbg("~p: firing action~n~n", [Op]),
            burm_action(Tree, Kids1, Rule)
    end.


%%%_. Error handling

%% burm_assert(Expr: bool) -> ok
%% returns `ok' iff its argument is true.
burm_assert(true)  -> ok;
burm_assert(false) -> burm_panic(assert).


%% burm_panic(Reason: error_descriptor)
%% throws a `burm_exn' exception.
burm_panic(Reason) -> throw({burm_exn, Reason}).


%% format_error(Reason: error_descriptor) -> string
%% returns a descriptive error message for `Reason'.
format_error(load) ->
    "error loading dumped tables";
format_error(dyncost) ->
    "burm_dyncost called with bad operator";
format_error({lookup, Table}) ->
    "bad result: " ++ atom_to_list(Table);
format_error(assert) ->
    "assertion failed";
format_error(noop) ->
    "no such operator";
format_error(nomatch) ->
    "subject tree cannot be matched".


%%%_. Debugging support
%%% Some custom reducers which print useful information about the result of
%%% labeling a subject tree.

%% dumpcover(Tree: node, Nt: int, Lvl: int) -> ok
%% prints the `Tree's rule cover. `Lvl' is the level of indentation.
dumpcover(Tree, Nt, Lvl) ->
    Rule = ?burm_rule(Nt, Tree),

    if Rule =:= ?INFINITY -> burm_panic(nomatch);
       true               -> ok
    end,

    Kids = kids(Tree),
    Nts  = burm_nts(Rule),
    case burm_is_chainrule(Rule) of
        true ->
            burm_print("~s~n", [burm_string({rule, Rule})], Lvl),
            dumpcover(Tree, hd(Nts), Lvl+1);
        false ->
            burm_print("~s~n", [burm_string({rule, Rule})], Lvl),
            [dumpcover(T, N, Lvl+1) || {T, N} <- lists:zip(Kids, Nts)]
    end,
    ok.


%% dumplabels(Tree: node) -> ok
%% prints `Tree's state labels.
dumplabels(Tree) ->
    io:put_chars("Operator: Nonterminal using Rule for Cost\n"),
    io:put_chars("=========================================\n"),
    dumplabels(Tree, 0),
    io:put_chars("\n").

dumplabels(Tree, Lvl) ->
    #burm_state{cost=C, rule=R} = state_label(Tree),
    burm_print("~p:~n", [op_label(Tree)], Lvl),
    [burm_print("~s using ``~s'' for ~p~n",
                [burm_string({nt, N}),
                 burm_string({rule, burm_rule(N, R)}),
                 element(N, C)],
                Lvl+1)
     || N <- lists:seq(1, size(C)),
        burm_rule(N, R) =/= ?INFINITY],
    [dumplabels(T, Lvl+2) || T <- kids(Tree)],
    ok.


%% burm_print(Fmt: string, Data: [term], N: int) -> ok
%% returns `ok' after `format'ting `Data' and indenting it `N' levels.
burm_print(Fmt, Data, Lvl) ->
    io:format(lists:duplicate(Lvl*2, $\s)++Fmt, Data).


%%%_. Emacs

%%% Local Variables:
%%% mode: erlang
%%% allout-layout: t
%%% time-stamp-pattern: "Modified : %%$"
%%% End:
