%%%============================================================================
%%% File     : eburg_backend.erl
%%% Purpose  : BURM emitter.
%%% Author   : j@svirfneblin.org
%%% License  : BSD
%%% Created  : 2007-09-21 11:56:51 j
%%% Modified : 2007-11-01 18:39:15 j
%%%============================================================================

%%%_. Attributes

-module(eburg_backend).
-author('j@svirfneblin.org').

-import(eburg_util,  [die/1, say/1, say/2, psay/1, ssay/2, sayforeach/3,
                      join/2, join/3, enumerate/1, gensym/1, pp/1, assoc/2]).
-import(eburg_rules, [is_chainrule/1, is_tainted/2, is_dynamic/1, is_dyncost/1,
                      mk_pvar_re/1, mk_rmap/1, irn2prn/3, walk_ltor/2,
                      collect_nts/1, arity/2, unparse_rule_short/1]).

-export([emit/1]).

-include("eburg.hrl").
-include("eburg_shared.hrl").


%%%_. Interface
%%% Most of BURM is printed directly to the output file using `io:format'.
%%% The convention is that `emit_*' routines print to ```get('$out_stream')'''
%%% (using `eburg_util:say') whereas `gen_*' routines return code (strings) for
%%% inclusion in some `emit_*' routine.
%%%
%%% The following optimizations from the iburg paper are currently implemented:
%%%   * inline record
%%%   * closure routines
%%%   * precompute leaf states
%%%   * pack rule numbers

%% emit(St: #eburg) -> ok
%% returns `ok'. Opens the output stream used by `say' and calls the top-level
%% `emit_*' routines.
%% uses the process dictionary.
%% dies with `file' if the output file couldn't be opened.
emit(St) ->
    case file:open(St#eburg.outfile, write) of
        {error, Reason} ->
            die({St#eburg.outfile, {file, Reason}});
        {ok, Stream} ->
            try
                put('$out_stream', Stream),
                say("%%% Generated by eburg, do not modify."),
                emit_preamble(St),
                emit_labeler(St),
                emit_reducer(St),
                say("%%% eof"),
                pp(St#eburg.outfile)
            after
                file:close(Stream),
                erase('$out_stream')
            end
    end.


%%%_. Preamble

emit_preamble(St) ->
    emit_dcls(St),
    emit_inn_macs(St),
    emit_user_code(St),
    emit_rewrite(St),
    emit_dumpcover(St),
    emit_dump(St),
    emit_load(St).


emit_dcls(St) ->
    say("-module(~s).", [St#eburg.module]),
    say("-author(eburg)."),
    say("-compile(nowarn_unused_vars)."),
    %% We need to export functions in burm.hrl here and include burm.hrl after
    %% the user code since that may contain export declarations as well.
    say("-export([rewrite/1, label/1, reduce/2, format_error/1,"),
    say("         dumpcover/1, dumplabels/1, dump/0]).").


emit_inn_macs(St) ->
    foreach(fun ({Nt, N}) -> say("-define(~s, ~p).", [Nt, N]) end,
            St#eburg.nt_to_inn).


emit_user_code(St) ->
    say("%%~s", [duplicate(77, $-)]),
    foreach(fun(Form) ->
                    psay(erl_prettypr:format(Form)),
                    psay("\n")
            end,
            St#eburg.code),
    say("%%~s", [duplicate(77, $-)]),
    say(""),
    say("-include(\"burm.hrl\").").


emit_rewrite(St) ->
    say("rewrite(Tree) -> reduce(label(Tree), ?~s).", [St#eburg.start]).


emit_dumpcover(St) ->
    say("dumpcover(Tree) ->"),
    say("    dumpcover(Tree, ?~s, 0),", [St#eburg.start]),
    say("    io:format(\"~n\").").


%% TODO: dump to single file
emit_dump(St) ->
    F = St#eburg.module,
    say("dump() ->"),
    say("    ets:tab2file(burm_id_cache, \"~s.ids\"),", [F]),
    say("    ets:tab2file(burm_state_cache, \"~s.states\"),", [F]),
    say("    ets:tab2file(burm_swdc_cache, \"~s.swdcs\").", [F]).


emit_load(St) ->
    Files = [St#eburg.module ++ Tab || Tab <- [".ids", ".states", ".swdcs"]],
    say("burm_maybe_load() ->"),
    say("    Results = [ets:file2tab(F) || F <- ~p],", [Files]),
    say("    not lists:member(error, [element(1, R) || R <- Results]).").


%%%_. Labeler

emit_labeler(St) ->
    emit_state(St),
    emit_closures(St),
    emit_packtabs(St),
    emit_dyncost(St),
    emit_is_tainted(St).


%%%_ , Leaves
%%% `static_leaves' finds those leaves which are static (none of the applicable
%%% rules depend on run-time information) and `sim_leaf' simulates the
%%% two-stage labeling process (recording baserules and computing the chainrule
%%% closure) at compile-compile time.
%%% Note that here we use tuples of tuples rather than #burm_state{} records
%%% to represent state labels.

%% static_leaves(St: #eburg) -> [atom]
%% returns those leaves whose rules all have static costs.
static_leaves(St) ->
    [L || L <- St#eburg.ops, arity(L, St) =:= 0, not is_tainted(L, St)].


%% sim_leaf(Rules: [#rule], St: #eburg) -> {tuple, tuple}
%% returns the state label for the leaf with baserules `Rules'.
sim_leaf(Rules, St) ->
    Init = mk_tuple(St),
    foldl(fun(Rule, State) -> sim_rule(Rule, State, St) end,
          {Init, Init},
          Rules).


%% mk_tuple(St: #eburg) -> tuple
%% returns a tuple which may be used to hold the cost or rule parts of a
%% state label.
mk_tuple(St) -> list_to_tuple(duplicate(length(St#eburg.nts), infinity())).


%% sim_rule(Rule: #rule, State: {tuple, tuple}, St: #eburg) -> {tuple, tuple}
%% returns the result of recording `Rule' in `State'.
sim_rule(Rule, State, St) ->
    %% base cost ist zero since leaves have no kids
    sim_record(Rule, State, 0, St).


%% sim_record(R: #rule,
%%            State: {tuple, tuple},
%%            B0: int,
%%            St: #eburg) -> {tuple, tuple}
%% returns the result of recording `R' in `State'.
sim_record(#rule{number=N, lhs=Nt, cost=Cost}, State={C0, R0}, B0, St) ->
    B1 = B0 + Cost, % known to be integer
    M  = assoc(Nt, St#eburg.nt_to_inn),
    case B1 < element(M, C0) of
        true ->
            C1 = setelement(M, C0, B1),
            R1 = setelement(M, R0, N),
            sim_closure(Nt, {C1, R1}, B1, St);
        false ->
            State
    end.


%% sim_closure(Nt: atom,
%%             State: {tuple, tuple},
%%             B0: int,
%%             St: #eburg) -> {tuple, tuple}
%% returns the closure of `State' under `Nt'.
sim_closure(Nt, State0, B0, St) ->
    foldl(fun (Rule, State) -> sim_record(Rule, State, B0, St) end,
          State0,
          assoc(Nt, St#eburg.crules)).


%%%_ , Rule packing
%%% Since we normalize the rule set we generate some additional nonterminals.
%%% `burm_state' copies a lot of state labels. Hence, implementing rule packing
%%% should save GC time.
%%% The idea is that instead of using tuples of integers (IRNs), we renumber
%%% the rules defining a single nonterminal internally and use only as many
%%% bits as we need for these packed rule numbers (PRNs).
%%% We generate functions for extracting and updating the bits used by each
%%% nonterminal from a binary and converting between IRNs and PRNs.

%% emit_packtabs(St: #eburg) -> ok
%% prints the tables used for rule packing.
emit_packtabs(St) ->
    {_Bits, Rmap} = mk_rmap(St),
    emit_getrule(Rmap),
    emit_decode(Rmap),
    emit_setrule(Rmap).


%% emit_getrule(Rmap: rmap) -> ok
%% prints one `burm_rule' clause for each nonterminal.
emit_getrule(Rmap) ->
    sayforeach(
      fun({Nt, _, _}) ->
              say("burm_rule(?~p, ~s) -> burm_decode_~p(~s)",
                  [Nt, gen_binary(Rmap), Nt, a2v(Nt)])
      end,
      ";",
      Rmap),
    say(".").


%% gen_binary(Rmap: rmap) -> string
%% returns code for the rule part of a state label.
gen_binary(Rmap)                    -> "<<" ++ gen_binary_loop(Rmap).
gen_binary_loop([{Nt, Bs, _}])      -> ssay("~s:~p>>", [a2v(Nt), Bs]);
gen_binary_loop([{Nt, Bs, _}|Rest]) ->
    ssay("~s:~p,", [a2v(Nt), Bs]) ++ gen_binary_loop(Rest).


%% a2v(Atom: atom) -> string
%% returns a variable name based on `Atom'.
a2v(Atom) ->
    Upcase = fun(X) when $a=<X, X=<$z -> X + $A-$a;
                (X)                   -> X
              end,
    [C|Cs] = atom_to_list(Atom),
    [Upcase(C)|Cs].


%% emit_decode(Rmap: rmap) -> ok
%% prints the PRN decoding tables.
emit_decode(Rmap) -> foreach(fun(X) -> emit_decode_loop(X) end, Rmap).

emit_decode_loop({Nt, _Bs, Rs}) ->
    sayforeach(
      fun({IRN, PRN}) -> say("burm_decode_~p(~p) -> ~p", [Nt, PRN, IRN]) end,
      ";",
      Rs),
    say(".").


%% emit_setrule(Rmap: rmap) -> ok
%% prints one `burm_setrule' clause for each nonterminal.
emit_setrule(Rmap) ->
    sayforeach(
      fun({Nt, _, _}) ->
              say("burm_setrule(?~p, ~s, Rule) -> ~s",
                  begin
                      Bin0 = gen_binary(Rmap),
                      case regexp:match(Bin0, "<"++a2v(Nt)++":") of
                          nomatch ->
                              {ok, Bin1, 1} =
                                  regexp:gsub(Bin0,","++a2v(Nt)++":",",Rule:");
                          _   ->
                              {ok, Bin1, 1} =
                                  regexp:gsub(Bin0, "<"++a2v(Nt)++":", "<Rule:")
                      end,
                      [Nt, Bin0, Bin1]
                  end)
      end,
      ";",
      Rmap),
    say(".").


%%%_ , Label calculation
%%% To find the optimal rule cover, BURM has to check, at each node, which
%%% nonterminals can be derived (and at what cost) using the baserules for
%%% the operator at that node and having done that whether any additional
%%% nonterminals can be derived using chainrules.
%%%
%%% Three things need to be done for each rule
%%%   (1) Check whether the rule matches the subtree rooted at the current node
%%%       (baserules only)
%%%   (2) Check whether the rule should be used (i.e., could some nonterminal
%%%       be derived more cheaply using the new rule?)
%%%   (3) If the rule should be used, record that fact in the state label
%%%
%%% The code to do (1)-(3) for the rules of a single operator is contained
%%% in the body of a clause of the case statement in `burm_state'. This code
%%% assumes that the initial cost and rule tuples are in `BurmIC' and `BurmIR';
%%% `burm_state' assumes that each case clause will put the final cost and rule
%%% tuples into `BurmFC' and `BurmFR'.

emit_state(St) ->
    say("burm_state(BurmNode) ->"),
    say("    BurmIC = ~p,", [mk_tuple(St)]),
    say("    BurmIR = ~s,", [gen_bin(St)]),
    say("    case op_label(BurmNode) of"),
    emit_cases(St),
    say("    end,"),
    say("    #burm_state{cost = BurmFC, rule = BurmFR}.").


%% gen_bin(St: #eburg) -> string
%% returns code for a zero-initialized binary.
gen_bin(St) ->
    {Bits, _Rmap} = mk_rmap(St),
    ssay("<<0:~p>>", [Bits]).


%% emit_cases(St: #eburg) -> ok
%% prints the body of the case statement in `burm_state'.
emit_cases(St) ->
    Ls = static_leaves(St),
    %% for each operator
    sayforeach(
      fun({Op, Rs}) ->
              say("~p ->", [Op]),

              case member(Op, Ls) of
                  true ->
                      %% static leaves' labels are precomputed
                      {C, R0} = sim_leaf(Rs, St),
                      R1      = gen_packed(R0, St),
                      say("{BurmFC, BurmFR} = {~p, ~s}", [C, R1]);
                  false ->
                      say("burm_assert(length(kids(BurmNode)) =:= ~p),",
                          [arity(Op, St)]),
                      %% for each of its rules
                      emit_clause_body(Rs, St)
              end
      end,
      ";",
      %% every op has baserules since we normalize the ruleset
      [{Op, assoc(Op, St#eburg.brules)} || Op <- St#eburg.ops]).


%% gen_packed(Rtup: {int}, St: #eburg) -> string
%% returns code for a binary containing the packed representation of `Rtup'.
gen_packed(Rtup, St) ->
    Rules         = tuple_to_list(Rtup),
    {_Bits, Rmap} = mk_rmap(St),
    "<<"++ join(fun({R, {_Nt, Bits, Rs}}) ->
                        ssay("~p:~p", [assoc(R, Rs), Bits])
                end,
                ",",
                zip(Rules, Rmap)) ++">>".


%%% Since intermediate versions of a node's state label have to be stored
%%% in different variables (and there may be arbitrarily many of these),
%%% we need some way of communicating the names of the variables holding
%%% the current of the cost and rule tuples between invocations of the
%%% generation routines.
%%%
%%% Currently, this is done by
%%%   * Passing the initial names as arguments
%%%   * Using `eburg_util:gensym' internally to generate new names
%%%   * Returning a tuple of the final names
%%%
%%% The relevant procedures are:
%%%   * `emit_body', which handles communication between the code for
%%%      different rules of a single operators
%%%   * `emit_record' and `emit_rule', which handle the code for a single rule

%% emit_clause_body(Rs: [#rule], St: #eburg) -> ok
%% prints the code for a single clause of the case statement in `burm_state'.
emit_clause_body(Rs, St) ->
    emit_body("{BurmFC, BurmFR} = {~s, ~s}",
              fun emit_rule/4,
              Rs, St, "BurmIC", "BurmIR").


%% emit_body(Final: string, Next: fun, Rs: [#rule], St: #eburg) -> ok
%% shared parts of `emit_clause_body' and `emit_closure_body'.
emit_body(Final, Next, [R|Rs], St, CurC, CurR) ->
    say("%% ~s", [unparse_rule_short(R)]),
    {NewC, NewR} = Next(R, St, CurC, CurR),
    say(","),
    emit_body(Final, Next, Rs, St, NewC, NewR);
emit_body(Final, _Next, [], _St, CurC, CurR) ->
    say(Final, [CurC, CurR]).


%%% Tree pattern matching and cost computation (c.f. (1) and (2) above) have
%%% been combined into a single test of rule applicability which inspects only
%%% the childrens' cost vectors with matching being done implicitly.
%%% Assume we are checking the rule
%%%
%%%   quux <- FOO(bar, baz)
%%%
%%% against the current node. Firstly, the current node's operator has to match
%%% FOO; this is ensured by the pattern part of the case clause.
%%% Secondly, it must be possible to reduce the left child to bar and the right
%%% child to baz. Since all elements of both the rule and cost vectors are
%%% initialized to a Very Large Integer, if we accept that no combination of
%%% rule applications will ever add up to a value as large as or larger than
%%% this, we can match the children by simply adding their cost vector entries
%%% for bar and baz respectively.
%%%
%%% `gen_cost' and `emit_record' implement this.

%% emit_rule(R: #rule,
%%           St: #eburg,
%%           CurC: string,
%%           CurR: string) -> {string, string}
%% returns the names of the variables holding the latest cost and rule vectors
%% after printing code for handling (matching & recording) a single rule.
emit_rule(R = #rule{rhs=#pattern{kids=Kids}}, St, CurC, CurR) ->
    B0 = gensym("Base"),
    say("begin"),
    say("    ~s = ~s,", [B0, if Kids =:= [] -> "0"; % leaf
                                true -> gen_cost(Kids)
                             end]),
    {NewC, NewR} = emit_record(R, St, CurC, CurR, B0),
    say("end"),
    {NewC, NewR}.


%% gen_cost(Kids: [atom]) -> string
%% returns code to sum the costs of deriving the nonterminals in `Kids'.
gen_cost(Kids) ->
    join(fun({Nt, N}) ->
                 io_lib:format("?burm_cost(?~s, lists:nth(~p, kids(BurmNode)))",
                               [Nt, N])
         end,
         " + ",
         enumerate(Kids)).


%% emit_record(Rule: #rule,
%%             St: #eburg,
%%             CurC: string,
%%             CurR: string,
%%             Base: string) -> {string, string}
%% prints the code for updating `CurC' and `CurR' according to `Rule'.
%% We have three cases to consider:
%%   * `Rule' is not applicable (too expensive)
%%   * `Rule' is applicable, and its LHS has no closure function
%%   * `Rule' is applicable, and its LHS does have a closure function
%% All of which have to result in the final rule and cost vectors residing
%% in the same variables, whose names are then returned to the caller.
emit_record(#rule{number=N, lhs=Nt, cost=Cost}, St, C0, R0, B0) ->
    case is_dyncost(Cost) of
        true  -> RC = subst_pseudo_vars(cost, Cost);
        false -> RC = integer_to_list(Cost)
    end,
    %% Fresh variable names
    B1 = gensym("Base"),
    C1 = gensym("Cost"),
    R1 = gensym("Rule"),
    %% Emit code
    say("~s = ~s + ~s,", [B1, B0, RC]),
    say("case ~s  < element(?~p, ~s) of", [B1, Nt, C0]),
    say("    true ->"),
    say("        ~s = setelement(?~p, ~s, ~s),", [C1, Nt, C0, B1]),
    say("        ~s = burm_setrule(?~p, ~s, ~p),",
        [R1, Nt, R0, irn2prn(Nt, N, St)]),
    %% Does `Nt' have a closure routine?
    case assoc(Nt, St#eburg.crules) of
        [] ->
            {NewC, NewR} = {C1, R1},
            say("{~s, ~s};", [C1, R1]);
        _ ->
            C2 = gensym("Cost"),
            R2 = gensym("Rule"),
            {NewC, NewR} = {C2, R2},
            say("{~s, ~s} = burm_closure_~s(BurmNode, ~s, ~s, ~s);",
                [C2, R2, Nt, C1, R1, B1])
    end,
    say("    false ->"),
    say("        {~s, ~s} = {~s, ~s}", [NewC, NewR, C0, R0]),
    say("end"),
    %% Return new variable names
    {NewC, NewR}.


%%% Each `burm_closure_*' function is much like a single clause in the case
%%% statement in `burm_state'.

%% emit_closures(St: #eburg) -> ok
%% prints one closure routine for every nonterminal which has chainrules.
emit_closures(St) ->
    %% for each nonterminal
    sayforeach(
      fun({Nt, Rs}) ->
              say("burm_closure_~s(BurmNode, BurmIC, BurmIR, BurmIB) ->", [Nt]),
              %% for each of its chainrules
              emit_closure_body(Rs, St)
      end,
      "",
      [{Nt, assoc(Nt, St#eburg.crules)} ||
          Nt <- St#eburg.nts, assoc(Nt, St#eburg.crules) =/= []]).


%% emit_closure_body(Rs: [#rule], St: #eburg) -> ok
%% prints the body of a single closure routine.
emit_closure_body(Rs, St0) ->
    emit_body("{~s, ~s}.",
              fun(R, St, CurC, CurR) ->
                      emit_record(R, St, CurC, CurR, "BurmIB")
              end,
              Rs, St0, "BurmIC", "BurmIR").


%%%_ , Memoization support
%%% The labeler in ```burm.hrl''' implements memoization, that is, it caches
%%% the result of calling `burm_state' on a node for possible re-use further
%%% up the tree (or in a different subject tree).
%%%
%%% The basic idea is that the result of `burm_state' depends only on the
%%% current node's operator, the state labels of its children, and on the
%%% results of evaluating the dynamic cost expressions of its rules in the
%%% context of the current node.
%%%
%%% We generate two functions which help with this: `burm_is_tainted' tells
%%% us whether an operator has any rules which have dynamic cost and if so,
%%% `burm_dyncost' returns a tuple holding the results of evaluating the dynamic
%%% cost expressions.

emit_is_tainted(St) ->
    sayforeach(
      fun(Op) -> say("burm_is_tainted(~p) -> ~p", [Op, is_tainted(Op, St)]) end,
      ";",
      St#eburg.ops),
    say("; burm_is_tainted(_) -> burm_panic(noop).").


emit_dyncost(St) ->
    Ops = [Op || Op <- St#eburg.ops, is_tainted(Op, St)],
    sayforeach(
      fun(Op) ->
              say("burm_dyncost(~p, BurmNode) -> ~s",
                  [Op, gen_dyncost(Op, St)])
      end,
      ";",
      Ops),
    if length(Ops) =/= 0 -> say(";");
       true -> ok
    end,
    say("burm_dyncost(_, _) -> burm_panic(dyncost).").


%% gen_dyncost(Op: atom, St: #eburg) -> string
%% returns code for a tuple whose elements are the dynamic cost expressions
%% occuring in rules which may be called when labeling a node whose root
%% operator is `Op'.
gen_dyncost(Op, St) ->
    Rs = walk_ltor(assoc(Op, St#eburg.brules), St#eburg.rules),
    Dyncosts = [subst_pseudo_vars(cost, R#rule.cost) || R <- Rs, is_dynamic(R)],
    "{" ++ join(", ", Dyncosts) ++ "}".


%%%_. Reducer
%%% eburg powered reducers rely on the following support functions:
%%%   * `burm_nts' which takes an IRN and returns the INNs of the nonterminals
%%%     occuring in that rule
%%%   * `burm_action' which is used to execute semantic action code
%%%   * `burm_is_chainrule' which does the obvious
%%%   * `burm_string' which maps IRNs and INNs to human-readable strings

emit_reducer(St) ->
    emit_nts(St),
    emit_action(St),
    emit_is_chainrule(St),
    emit_string(St).


emit_nts(St) ->
    sayforeach(
      fun(R) -> say("burm_nts(~p) -> ~s", [R#rule.number, gen_nts(R)]) end,
      ";",
      St#eburg.rules),
    say(".").


%% gen_nts(R: #rule, St: #eburg) -> string
%% returns code for a list containing the INNs of the nonterminals on the
%% RHS of `R' in the order in which they appear on the RHS of `R'.
gen_nts(R) ->
    io_lib:format("[ ~s ]",
                  [join(fun(Nt) -> io_lib:format("?~p", [Nt]) end,
                        ", ",
                        collect_nts(R#rule.rhs))]).


%% emit_action(St: #eburg) -> ok
%% prints the semantic actions, encapsulated in a separate routine.
emit_action(St) ->
    say("burm_action(BurmNode, BurmReducedKids, BurmRule) ->"),
    say("    case BurmRule of"),
    emit_actions(St),
    say("    end.").


%% emit_actions(St: #eburg) -> ok
%% print the body of the case statement in `burm_action'.
emit_actions(St) ->
    sayforeach(
      fun(R) ->
              say("~p ->", [R#rule.number]),
              case R#rule.action of
                  ""   -> psay("ok");
                  Code -> psay(subst_pseudo_vars(action, Code))
              end
      end,
      ";",
      St#eburg.rules).


emit_is_chainrule(St) ->
    sayforeach(
      fun(R) ->
              say("burm_is_chainrule(~p) -> ~p",
                  [R#rule.number, is_chainrule(R)])
      end,
      ";",
      St#eburg.rules),
    say(".").


emit_string(St) ->
    %% Rule names
    sayforeach(
      fun(R) ->
              say("burm_string({rule, ~p}) -> ~p",
                  [R#rule.number, unparse_rule_short(R)])
      end,
      ";",
      St#eburg.rules),
    say(";"),
    %% Nonterminal names
    sayforeach(
      fun (Nt) ->
              say("burm_string({nt, ~p}) -> ~p",
                  [assoc(Nt, St#eburg.nt_to_inn), atom_to_list(Nt)])
      end,
      ";",
      St#eburg.nts),
    say(".").


%%%_. Pseudo variables
%%% Pseudo variables can occur in two places: Dynamic costs and semantic
%%% actions.
%%% Dynamic cost code is included in `burm_state', `closure_*', and
%%% `burm_dyncost'. Semantic action code is included in `burm_action'.
%%%
%%% Semantic action code may refer to the current node as `$a' and to
%%% the results of reducing the current node's children as `$1' to `$9'
%%% (more specifically, `$N' is the result of reducing the nth nonterminal
%%% on the RHS of the rule which is being used to reduce the current node).
%%% Dynamic cost code can only refer to the current node (also as `$a').
%%%
%%% The below code assumes that the current node is pointed to by the
%%% variable `BurmNode'.

%% subst_pseudo_vars(Tag: atom, Code: string) -> string
%% returns `Tag' string `Code' with pseudo variables expanded.
subst_pseudo_vars(cost, Cost)     -> subst_node(Cost);
subst_pseudo_vars(action, Action) -> subst_kids(subst_node(Action)).


%% subst_node(Code: string) -> string
%% substitute any references to '$a' in `Code' with ```BurmNode'''.
subst_node(Code0) ->
    {ok, Code1, _} = regexp:gsub(Code0, mk_pvar_re(a), "BurmNode"),
    Code1.


%% subst_kids(Code: string) -> string
%% substitute references to '$N', N in [1-9], with
%% ```lists:nth(N, BurmRedcedKids)'''.
subst_kids(Code) ->
    foldl(fun(N, Code0) ->
                  {ok, Code1, _} =
                      regexp:gsub(Code0,
                                  mk_pvar_re(N),
                                  "lists:nth(" ++ integer_to_list(N)
                                  ++ ", BurmReducedKids)"),
                  Code1
          end,
          Code,
          seq(1, 9)).


%%%_. Emacs
%%% Local Variables:
%%% mode: erlang
%%% allout-layout: t
%%% time-stamp-pattern: "Modified : %%$"
%%% End:
