%%%============================================================================
%%% File     : eburg_frontend.erl
%%% Purpose  : EBURG front end.
%%% Author   : j@svirfneblin.org
%%% License  : BSD
%%% Created  : 2007-09-21 11:55:04 j
%%% Modified : 2007-11-01 17:21:18 j
%%%============================================================================

%%%_. Attributes

-module(eburg_frontend).
-author('j@svirfneblin.org').

-import(eburg_util,  [die/1, maybe_die/2, join/3, uniq/1, enumerate/1,
                      gensym/1, mk_reader/2]).
-import(eburg_rules, [is_dynamic/1, mk_pvar_str/1, mk_pvar_re/1, patflatmap/2,
                      walk_rtol/3, count_nts/1, dumprules/1]).

-export([read/2]).

-include("eburg.hrl").
-include("eburg_shared.hrl").


%%%_. Interface

%% read(File: name, Opts: proplist) -> #eburg
%% returns the state record from which to generate BURM.
read(File, Opts) ->
    St0 = check(reparse1(init(File, Opts))),
    ?verbose(begin
                 io:put_chars("Initial rule set:\n"),
                 dumprules(St0#eburg.rules)
             end),
    St1 = reparse3(reparse2(normalize(St0))),
    ?verbose(begin
                 io:put_chars("\n\nFinal rule set:\n"),
                 dumprules(St1#eburg.rules)
             end),
    St1.


%%%_. Initialization
%%% The procedure `init' (after processing options) constructs the initial
%%% #eburg record. This state record has:
%%%   * The `ops' field set (and its value is final)
%%%   * The `start' field set (likewise final)
%%%   * The `rules' field set, but this field's value is not final:
%%%       (a) The rules' `number' field is not yet set.
%%%       (b) Normalization will likely introduce new rules and nonterminals.
%%%       (c) Dyncosts need to be wrapped in a catch.
%%%   * All other fields are unset
%%% `init' also sets the `infile', `outfile', `module', and `code' fields
%%% to their final values.

%% init(File: name, Opts: proplist) -> #eburg
%% uses the process dictionary.
init(File, Opts) ->
    InDir   = filename:dirname(File),
    OutDir  = proplists:get_value(outdir, Opts, InDir),
    Module  = filename:basename(File, ".brl"),
    InFile  = InDir  ++ "/" ++ Module ++ ".brl",
    OutFile = OutDir ++ "/" ++ Module ++ ".erl",
    Verbose = proplists:get_value(verbose, Opts, false),
    put('$verbose', Verbose),

    (parse(InFile))#eburg{infile  = InFile,
                          outfile = OutFile,
                          module  = Module}.


%% parse(File: string) -> #eburg
%% returns the internal representation of the specification in `File'.
%% dies with `file' if `File' couldn't be opened, `syntax' if the
%% specification contains syntax errors, `code' if the user code can't
%% be parsed.
parse(File) ->
    case file:open(File, read) of
        {error, Reason} ->
            die({File, {file, Reason}});
        {ok, Stream} ->
            try
                {Spec, Code} = get_file(Stream),
                Errs = [{File, Line, {code, Mod, Descr}} ||
                           {error, {Line, Mod, Descr}} <-
                               [F || F <- Code, element(1, F) =:= error]],
                maybe_die(Errs, []),
                Reader = mk_reader(fun eburg_lexer:lex/1,
                                   fun eburg_parser:parse/1),
                case Reader(Spec) of
                    {error, {Line, Mod, Descr}} ->
                        die({File, Line, {syntax, Mod, Descr}});
                    St ->
                        St#eburg{code=Code}
                end
             after
                 file:close(Stream)
             end
    end.


%% get_file(File: io_device) -> {string, [syntaxTree]}
%% returns the contents of `File', split into declarations & rules
%% and user code.
get_file(Stream) ->
    get_file(Stream, [], 1).

get_file(Stream, Lines, N) ->
    case io:get_line(Stream, '') of
        eof ->
            {Lines, []};
        "Erlang code.\n" ->
            {ok, Forms} = epp_dodger:parse(Stream, N+1),
            {Lines, Forms};
        Line ->
            get_file(Stream, Lines ++ Line, N+1)
    end.


%%%_. Basic input validation
%%% The procedure `check' tests for the following:
%%%   * syntactic correctness of semantic actions and dyncosts (using stdlib
%%%     facilities)
%%%   * sanity of declarations
%%%   * validity of tree patterns
%%%   * are all operators used, all nonterminals reachable?

%% check(St: #eburg) -> #eburg
%% returns its input unchanged if it is found to be valid.
%% dies with `action', `cost', `dup', `start', `undeclared', and/or `arity'
%% if the corresponding `check_*' routine returns errors.
%% warns about `unused' and `unreachable'.
check(St) ->
    Collect = fun(Test, Errs) -> Errs ++ Test(St) end,

    Errs =
        foldl(Collect, [],
              [fun check_code/1,
               fun check_dup/1,
               fun check_start/1,
               fun check_undeclared/1,
               fun check_arity/1]),

    Warns =
        foldl(Collect, [],
              [fun check_unused/1,
               fun check_unreachable/1]),

    maybe_die(Errs, Warns),
    St.


%% check_code(St: #eburg) -> [error_info]
%% returns syntax errors in dynamic cost and semantic action code.
check_code(St) ->
    Rs      = St#eburg.rules,
    Reader  = mk_reader(fun erl_scan:string/1, fun erl_parse:parse_exprs/1),
    Actions = [{R, R#rule.action++"."} || R <- Rs, R#rule.action =/= ""],
    Dcosts  = [{R, R#rule.cost++"."} || R <- Rs, is_dynamic(R)],
    Collect = fun(Tag) ->
                      fun({R, Code}, Errs) ->
                              case Reader(Code) of
                                  {error, {_Line, Mod, Descr}} ->
                                      Errs ++ [{St#eburg.infile,
                                                R#rule.line,
                                                {Tag, Mod, Descr}}];
                                  _ -> Errs
                              end
                      end
              end,
    foldl(Collect(action), [], Actions) ++ foldl(Collect(cost), [], Dcosts).


%% check_dup(St: #eburg) -> [error_info]
%% returns an error if there are multiply declared operators.
check_dup(St) ->
    Ops  = St#eburg.ops,
    Dups = uniq(Ops -- uniq(Ops)),
    if Dups =/= [] -> [{St#eburg.infile, {dup, Dups}}];
       true        -> []
    end.


%% check_start(St: #eburg) -> [error_info]
%% returns an error if there are no rules for the start symbol.
check_start(St) ->
    case not member(St#eburg.start, St#eburg.nts) of
        true  -> [{St#eburg.infile, start}];
        false -> []
    end.


%% check_undeclared(St: #eburg) -> [error_info]
%% returns errors for operators which haven't been declared
check_undeclared(St) ->
    Undeclared =
        flatten(
          map(fun(R) ->
                      patflatmap(fun(#pattern{op=Op}) ->
                                         case member(Op, St#eburg.ops) of
                                             true  -> [];
                                             false -> [{Op, R#rule.line}]
                                         end
                                 end,
                                 [R#rule.rhs])
              end,
              St#eburg.rules)),
    [{St#eburg.infile, L, {undeclared, Op}} || {Op, L} <- Undeclared].


%% check_arity(St: #eburg) -> [error_info]
%% returns errors for operators used with different arities.
check_arity(St) ->
    UnSingleAritied =
        [{Op, Ns} ||
            {Op, Ns} <- [{Op, lists:usort(proplists:get_all_values(
                                      Op,
                                      patflatmap(fun(#pattern{op=O, kids=K}) ->
                                                         [{O, length(K)}]
                                                 end,
                                                 [R#rule.rhs ||
                                                     R <- St#eburg.rules])))} ||
                            Op <- St#eburg.ops],
                     length(Ns) > 1],
    lists:usort(
      [{St#eburg.infile, {arity, Op, Ns}} || {Op, Ns} <- UnSingleAritied]).


%% check_unused(St: #eburg) -> [error_info]
%% returns warnings for any unused operators.
check_unused(St) ->
    Used   = patflatmap(fun(#pattern{op=Op}) -> [Op] end,
                        [R#rule.rhs || R <- St#eburg.rules]),
    Unused = St#eburg.ops -- Used,
    [{St#eburg.infile, {unused, Op}} || Op <- Unused].


%% check_unreachable(St: #eburg) -> [error_info]
%% returns warnings for any unreachable nonterminals.
check_unreachable(St) ->
    Rs     = St#eburg.rules,
    Start  = St#eburg.start,
    Reach  = walk_rtol([R || R <- Rs, R#rule.lhs =:= Start], [Start], Rs),
    Seen   = [R#rule.lhs || R <- Reach],
    Unseen = St#eburg.nts -- Seen,
    [{St#eburg.infile, {unreachable, Nt}} || Nt <- Unseen].


%%%_. Normalization
%%% To simplify the implementation of state equivalence, reducers, &c.
%%% we normalize the input grammar.
%%% A rule is in normal form if it is either a chainrule or an baserule whose
%%% RHS operator's children are all nonterminals (i.e., rule normalization
%%% eliminates nested patterns).
%%%
%%% An example to illustrate the process:
%%% The rule set
%%%
%%%     foo <- BAR(baz, QUUX(FROB(blorp), snarfle)) : ['$1', '$2', '$3'];
%%%
%%% is not in normal form, since QUUX(...) is not a nonterminal.
%%% A single normalization step yields:
%%%
%%%     foo <- BAR(baz, burm_nt0) : ['$1', element(1, '$2'), element(2, '$2')];
%%%     burm_nt0 <- QUUX(FROB(blorp), snarfle) : {'$1', '$2'};
%%%
%%% Notice how both the rule's RHS and its semantic action have to be
%%% rewritten. A second normalization step gives the final rule set:
%%%
%%%     foo <- BAR(baz, burm_nt0) : ['$1', element(1, '$2'), element(2, '$2')];
%%%     burm_nt0 <- QUUX(burm_nt1, snarfle) : {'$1', '$2'};
%%%     burm_nt1 <- FROB(blorp) : '$1';
%%%
%%% Two "optimizations":
%%%   * Reuse rules whenever possible
%%%   * The semantic actions of newly introduced rules with only a single
%%%     nonterminal on their RHS return its value directly (rather than
%%%     wrapping it in a tuple)


%% normalize(St: #eburg) -> #eburg
%% sets up the `rule_cache' table and calls `do_normalize'.
normalize(St) ->
    %% for interactive development
    case member(rule_cache, ets:all()) of
        true  -> ets:delete(rule_cache);
        false -> ok
    end,
    ets:new(rule_cache, [named_table]), % :: {RHS: #pattern, LHS: atom}
    St#eburg{rules=do_normalize(St#eburg.rules)}.


%% do_normalize(Rs: [#rule]) -> [#rule]
%% iterates over `Rs' until it becomes stable.
do_normalize(Rs0) ->
    {Changes, Rs1} = normalize_loop(Rs0),
    if Changes -> do_normalize(Rs1);
       true    -> Rs1
    end.


%% normalize_loop(Rs: [#rule]) -> [#rule]
%% loops over `Rs' calling `normalize_rule' as necessary.
%% We try to keep the rules in the order in which they appeared in the
%% input file, with new rules spliced in after their parent rules
%% (ordered by child replaced).
normalize_loop(Rs) ->
    normalize_loop(Rs, [], false).

normalize_loop([], Rules, Flag) ->
    {Flag, Rules};
normalize_loop([R|Rs], Rules, Flag) ->
    case is_normalized(R) of
        true  -> normalize_loop(Rs, Rules ++ [R], Flag);
        false -> normalize_loop(Rs, Rules ++ normalize_rule(R), true)
    end.


%% is_normalized(R: #rule) -> bool
%% returns `true' if `R' is in normal form.
is_normalized(#rule{rhs=Nt}) when is_atom(Nt) ->
    true;
is_normalized(#rule{rhs=#pattern{kids=Kids}}) ->
    all(fun erlang:is_atom/1, Kids).


%% normalize_rule(R: #rule) -> [#rule]
%% returns `R' in normal form and new rules for the sub-patterns on
%% `R's RHS. `R' is assumed to _not_ be in normal form already.
normalize_rule(R=#rule{rhs=#pattern{kids=Kids0}}) ->
    {Kids1, Rs} = rewrite_pattern(Kids0),
    Ac = rewrite_action(R),
    [R#rule{rhs=(R#rule.rhs)#pattern{kids=Kids1}, action=Ac}] ++ Rs.


%% rewrite_pattern(Kids: [atom|#pattern]) -> {[atom], [#rule]}
%% returns the new list of kids (now all atomic) and a list of new rules
%% (which may have to be normalized further). Also enters the new rules
%% into the `rule_cache' table (in fact, it only puts the new rules'
%% LHS, indexed by original sub-pattern, into the ETS table; this rule
%% may be further normalized later but its LHS will remain the same).
%% New rules have zero cost.
rewrite_pattern(Kids) ->
    rewrite_pattern(Kids, [], []).

rewrite_pattern([], Kids, Rules) ->
    {reverse(Kids), Rules};
rewrite_pattern([K|Ks], Kids, Rules) when is_atom(K) ->
    rewrite_pattern(Ks, [K|Kids], Rules);
rewrite_pattern([K|Ks], Kids, Rules) when is_record(K, pattern)->
    case ets:lookup(rule_cache, K) of
        [{_, Nt}] ->
            rewrite_pattern(Ks, [Nt|Kids], Rules);
        [] ->
            Nt = gensym(burm_nt),
            ets:insert(rule_cache, {K, Nt}),
            R = #rule{lhs=Nt, rhs=K, action=gen_action(K)},
            rewrite_pattern(Ks, [Nt|Kids], Rules ++ [R])
    end.


%% gen_action(Pat: #pattern) -> string
%% returns the semantic action code for a generated rule with RHS `Pat'.
%% The rule will return the values its nonterminals reduce to as an N-tuple
%% where N is the number of these "free variables".
gen_action(Pat) ->
    case count_nts(Pat) of
        0 -> "ok";
        1 -> mk_pvar_str(1);
        N -> "{"++ join(fun eburg_rules:mk_pvar_str/1, ", ", seq(1, N)) ++"}"
    end.


%% rewrite_action(R: #rule) -> #rule
%% returns `R' with pvar references in its semantic action code rewritten
%% to match the normalized rule set.
rewrite_action(#rule{rhs=#pattern{kids=Kids}, action=A}) ->
    Pass1 =
        foldl(fun({Old, New}, Ac0) when is_integer(New) ->
                      {ok, Ac1, _} = regexp:gsub(Ac0,
                                                 mk_pvar_re(Old),
                                                 "burm_temp"++
                                                 integer_to_list(New)),
                      Ac1;
                 ({Old, {elt, Pos, New}}, Ac0) ->
                      {ok, Ac1, _} = regexp:gsub(Ac0,
                                                 mk_pvar_re(Old),
                                                 "element(" ++
                                                 integer_to_list(Pos) ++ ", "++
                                                 "burm_temp"++
                                                 integer_to_list(New)++ ")"),
                      Ac1
              end,
              A,
              get_pvar_mapping(Kids)),
    foldl(fun(N, Acc0) ->
                  {ok, Acc1, _} = regexp:gsub(Acc0,
                                              "burm_temp"++integer_to_list(N),
                                              mk_pvar_str(N)),
                  Acc1
          end,
          Pass1,
          seq(1, 9)).


%% get_pvar_mapping(Kids: [atom|#pattern]) -> [{int,int}|{int,{elt,int,int}}]
%% returns the mapping of old pvars to new (post-normalization) pvars.
%% A value of {N, M} means that the old pvar number N maps to the new pvar M,
%% whereas a value of {N, {elt, Pos, M}} means that the old pvar N maps to
%% the Pos-th element of the new pvar M.
get_pvar_mapping(Kids) ->
    get_pvar_mapping(Kids, 1, 1).

get_pvar_mapping([K|Ks], N, M) when is_atom(K) ->
    [{N, M}|get_pvar_mapping(Ks, N+1, M+1)];
get_pvar_mapping([K|Ks], N0, M) when is_record(K, pattern) ->
    X = count_nts(K),
    if X=:=0 -> get_pvar_mapping(Ks,N0,M+1); % skip
       X=:=1 -> [{N0, M}|get_pvar_mapping(Ks, N0+1, M+1)];
       %% new pvar M returns old pvars N..N+X-1
       true -> Map = [{N1, {elt, Pos, M}} ||
                         {N1, Pos} <- zip(seq(N0, N0+X-1), seq(1, X))],
               Map ++ get_pvar_mapping(Ks, N0+X, M+1)
    end;
get_pvar_mapping([], _N, _M) ->
    [].


%%%_. Reparsing
%%% Here we
%%%   (a) attach Internal Rule Numbers (IRNs) to rules
%%% and fill in the remaining state record fields:
%%%   (b) `nts', the known nonterminals
%%%   (c) `nt_to_inn', mapping of nonterminals to Internal Nonterminal Numbers
%%%   (d) `crules', chainrules indexed by RHS nonterminal
%%%   (e) `brules', baserules indexed by RHS operator
%%% INNs and IRNs number nonterminals/rules _in the order in which they appear
%%% in the spec file_ (modulo normalization).
%%% Nonterminals are defined by their occurence on the LHS of some rule.
%%% Finally, we also
%%%   (f) drop unused operators
%%%   (g) dynamic cost computations are wrapped in catch begin ... end.

%% reparse1(St: #eburg) -> #eburg
%% does (b) and (g). Should be called before `check'.
reparse1(St) ->
    Rs0 = St#eburg.rules,
    Nts = uniq([R#rule.lhs || R <- Rs0]),
    Rs1 = map(
            fun(R) ->
                    case is_dynamic(R) of
                        true ->
                            R#rule{cost=
                 begin
                     BurmTemp = gensym("BurmTemp"),
                     "begin "++
                          BurmTemp++" = (catch begin "++R#rule.cost++" end), "++
                          "case is_integer("++BurmTemp++") of "++
                              "true  -> "++BurmTemp++"; "++
                              "false -> "++integer_to_list(infinity())++" "++
                          "end "++
                     "end"
                 end};
                        false -> R
                    end
            end,
            Rs0),
    St#eburg{nts   = Nts,
             rules = Rs1}.


%% reparse2(St: #eburg) -> #eburg
%% does (f). Call after `check'.
reparse2(St) ->
    Ops = patflatmap(fun(#pattern{op=Op}) -> [Op] end,
                     [R#rule.rhs || R <- St#eburg.rules]),
    St#eburg{ops=uniq(Ops)}.


%% reparse3(St: #eburg) -> #eburg
%% does (a) to (e). Should be called after `normalize'.
reparse3(St) ->
    Rs  = [R#rule{number=N} || {R, N} <- enumerate(St#eburg.rules)],
    Nts = uniq([R#rule.lhs || R <- Rs]),
    St#eburg{nts       = Nts,
             nt_to_inn = enumerate(Nts),
             rules     = Rs,
             crules    = [{Nt, [R || R <- Rs, R#rule.rhs=:=Nt]} || Nt <- Nts],
             brules    = [{Op, [R || R <- Rs, (R#rule.rhs)#pattern.op=:=Op]} ||
                             Op <- St#eburg.ops]}.


%%%_. Emacs
%%% Local Variables:
%%% mode: erlang
%%% allout-layout: t
%%% time-stamp-pattern: "Modified : %%$"
%%% End:
