%%%============================================================================
%%% File     : eburg_rules.erl
%%% Purpose  : Functions for working with rules and patterns.
%%% Author   : j@svirfneblin.org
%%% License  : BSD
%%% Created  : 2007-09-20 13:00:30 j
%%% Modified : 2007-11-01 17:26:05 j
%%%============================================================================

%%%_. Attributes

-module(eburg_rules).
-author('j@svirfneblin.org').

-import(eburg_util, [assoc/2, next_pot/1, next_moe/1, enumerate/1]).

-export([is_chainrule/1, is_tainted/2, is_dynamic/1, is_dyncost/1]).
-export([mk_pvar_re/1, mk_pvar_str/1]).
-export([mk_rmap/1, irn2prn/3]).
-export([patmap/2, patflatmap/2]).
-export([walk_rtol/3, walk_ltor/2]).
-export([collect_nts/1, count_nts/1, arity/2]).
-export([dumprules/1, unparse_rule_short/1]).

-include("eburg.hrl").
-include("eburg_shared.hrl").


%%%_. Type tests

%% is_chainrule(R: #rule) -> bool
is_chainrule(#rule{rhs=Nt})  when is_atom(Nt)             -> true;
is_chainrule(#rule{rhs=Pat}) when is_record(Pat, pattern) -> false.


%% is_baserule(R: #rule) -> bool
%% is_baserule(R) -> not is_chainrule(R).


%% is_tainted(Op: atom, St: #eburg) -> bool
%% returns `true' if `Op' is tainted.
%% Operators are tainted if any of their baserules or any of the chainrules
%% which occur in the closures of their baserules are dynamic.
is_tainted(Op, St) ->
    Rs = assoc(Op, St#eburg.brules),
    any(fun is_dynamic/1, Rs) orelse
        [R || R <- walk_ltor(Rs, St#eburg.rules), is_dynamic(R)] =/= [].


%% is_dynamic(R: #rule) -> bool
is_dynamic(#rule{cost=C}) -> is_dyncost(C).


%% is_dyncost(Cost: string|int) -> bool
%% returns `true' if `Cost' is dynamic (i.e. an Erlang expression rather than
%% an integer constant).
is_dyncost(Cost) when is_list(Cost)    -> true;
is_dyncost(Cost) when is_integer(Cost) -> false.


%%%_. Pseudo variable constructors
%%% Known pseudo variables: '$a', '$1' ... '$9'.
%%% We need them as strings when generating semantic actions in the front end
%%% and as regexps when replacing them with the actual variable names.

%% mk_pvar_re(Pvar: atom|int) -> regexp
%% returns a regexp for matching `Pvar'.
mk_pvar_re(Pvar) -> "'\\$" ++ concat([Pvar]) ++ "'".


%% mk_pvar_str(Pvar: atom|int) -> string
%% returns the string name of `Pvar'.
mk_pvar_str(Pvar) -> "'$" ++ concat([Pvar]) ++ "'".


%%%_. Rule packing support
%%% See `eburg_backend'.

%% mk_rmap(St: #eburg) -> rmap
%%   where
%%     rmap = {int, [{atom, int, [{int, int}]}]}
%% returns the total number of bits required for the rule vector and
%% associates with each nonterminal
%%   * the number of bits required to count its rules
%%   * a mapping of IRNs to PRNs
%% A PRN equal to zero indicates ``no match''.
mk_rmap(St) ->
    Rmap0 = map(fun(Nt) ->
                        Rs = defined_by(Nt, St),
                        {Nt,
                         next_pot(length(Rs)),
                         [{infinity(), 0}] ++
                         enumerate([R#rule.number || R <- Rs])}
                end,
                St#eburg.nts),
    Bits  = next_moe(lists:sum([Bs || {_, Bs, _} <- Rmap0])),
    Rmap1 = pad_last(Rmap0, Bits),
    {Bits, Rmap1}.


%% defined_by(Nt: atom, St: #eburg) -> [#rule]
%% returns the rules which define `Nt'.
defined_by(Nt, St) -> [R || R <- St#eburg.rules, R#rule.lhs =:= Nt].


%% pad_last(Rmap: rmap, Bits: int) -> rmap
%% since the size of Erlang binaries has to be divisible by eight, we might have
%% to use more bits than strictly necessary for the last nonterminal.
pad_last([{Nt, _Bs, Rs}], Bits)   -> [{Nt, Bits, Rs}];
pad_last([X={_, Bs, _}|Xs], Bits) -> [X|pad_last(Xs, Bits-Bs)].


%% irn2prn(Nt: aton, IRN: int, St: #eburg) -> int
%% returns the PRN corresponding to `IRN' under `St'.
irn2prn(Nt, IRN, St) ->
    {_Bits, Rmap}        = mk_rmap(St),
    {value, {Nt, _, Rs}} = lists:keysearch(Nt, 1, Rmap),
    assoc(IRN, Rs).    


%%%_. Mapping
%%% `patmap' may be used for functional update of an RHS.
%%% `patflatmap' is useful when one wants to extract something
%%% from an RHS.

%% patmap(F: fun, RHS: [atom|#pattern]) -> [term]
%% iterates over `RHS', replacing any #pattern's it comes across with the
%% result of calling `F' on them. Children are traversed depth-first.
%% Nonterminals (atoms) are included in the result unchanged.
patmap(F, [P = #pattern{kids=Kids0}|Rest]) ->
    Kids1 = patmap(F, Kids0),
    [F(P#pattern{kids=Kids1})|patmap(F, Rest)];
patmap(F, [Nt|Rest]) when is_atom(Nt) ->
    [Nt|patmap(F, Rest)];
patmap(F, []) when is_function(F, 1) -> [].


%% patflatmap(F: fun, RHS: [atom|#pattern]) -> [term]
%% traverses `RHS', discarding any atoms and concatenating the results
%% of calling `F' on any patterns. `F' must return a list.
patflatmap(F, [P = #pattern{kids=Kids}|Rest]) ->
    F(P) ++ patflatmap(F, Kids) ++ patflatmap(F, Rest);
patflatmap(F, [Nt|Rest]) when is_atom(Nt) ->
    patflatmap(F, Rest);
patflatmap(F, []) when is_function(F, 1) -> [].


%%%_. Traversing the rule set
%%% `walk_ltor' and `walk_rtol' view the input grammar as a graph:
%%% Rules are nodes and rule B is reachable from rule A either
%%% if rule B's LHS occurs on rule A's RHS (right-to-left) or
%%% if rule A's LHS is equal to rule B's RHS (left-to-right).

%% walk_rtol(Rs: [#rule], Seen: [atom], Ruleset: [#rule]) -> [#rule]
%% returns all elements of `Ruleset' which are rtol-reachable from
%% any of the rules in `Rs'.
walk_rtol([R = #rule{rhs=Nt}|Rs], Seen, Ruleset) when is_atom(Nt) ->
    NewRules =
        case not member(Nt, Seen) of
            true  -> [Rule || Rule <- Ruleset, Rule#rule.lhs =:= Nt];
            false -> []
        end,
    [R] ++ walk_rtol(Rs ++ NewRules, [Nt|Seen], Ruleset);
walk_rtol([R = #rule{rhs=Pat}|Rs], Seen, Ruleset) when is_record(Pat, pattern)->
    Nts      = collect_nts(R#rule.rhs),
    NewRules = [Rule || Rule <- Ruleset,
                        member(Rule#rule.lhs, Nts -- Seen)],
    [R] ++ walk_rtol(Rs ++ NewRules, Nts ++ Seen, Ruleset);
walk_rtol([], _Seen, _Ruleset) -> [].


%% walk_ltor(Rs: [#rule], Ruleset: [#rule]) -> [#rule]
%% returns all elements of `Ruleset' which are ltor-reachable from
%% any of the rules in `Rs'.
walk_ltor(Rs, Ruleset) ->
    walk_ltor(Rs, [], Ruleset).

walk_ltor([R = #rule{lhs=Nt}|Rs], Seen, Ruleset) ->
    NewRules = 
        case not member(Nt, Seen) of
            true  -> [Rule || Rule <- Ruleset, Rule#rule.rhs =:= Nt];
            false -> []
        end,
    [R] ++ walk_ltor(Rs ++ NewRules, [Nt|Seen], Ruleset);
walk_ltor([], _Seen, _Ruleset) -> [].


%%%_. Analyzing RHSs

%% collect_nts(RHS: #pattern|atom) -> [atom]
%% returns the nonterminals occuring in `RHS'.
collect_nts(Nt) when is_atom(Nt) -> [Nt];
collect_nts(Pat) when is_record(Pat, pattern) ->
    [Nt || Nt <- flatten(patmap(fun(#pattern{kids=Kids}) -> Kids end, [Pat]))].


%% count_nts(RHS: #pattern|atom) -> int
%% returns the number of nonterminals occuring in `RHS'.
count_nts(Nt) when is_atom(Nt) -> 1;
count_nts(#pattern{kids=Kids}) -> lists:sum(map(fun count_nts/1, Kids)).


%% arity(Op: atom, St: #eburg]) -> int
%% returns the arity of `Op' as used in `St's rules.
arity(Op, St) ->
    [#rule{rhs=#pattern{op=Op, kids=Kids}}|_Rest] = assoc(Op, St#eburg.brules),
    length(Kids).


%%%_. Debugging helpers

%% dumprules(Rs: [#rule]) -> ok
%% prints `Rs' to stdout.
dumprules(Rs) ->
    foreach(fun(R) -> io:put_chars(unparse_rule(R)) end, Rs).


%% unparse_rule(R: #rule) -> string
%% returns the full string representation of `R'.
unparse_rule(#rule{number=N, lhs=LHS, rhs=RHS, cost=C, action=A}) ->
    io_lib:format("~p. ~p <- ~p [~p] :", [N, LHS, unparse_rhs(RHS), C])
        ++ A ++ "\n".


%% unparse_rule_short(R: #rule) -> string
%% returns the short string representation of `R'.
unparse_rule_short(#rule{lhs=LHS, rhs=RHS}) ->
    io_lib:format("~p <- ~s", [LHS, unparse_rhs(RHS)]).


%% unparse_rhs(RHS: [atom|#pattern]|atom|#pattern) -> string
%% returns the string representation of `RHS'.
unparse_rhs([]) ->
    "";
unparse_rhs([X]) ->
    unparse_rhs(X);
unparse_rhs([X|Xs]) ->
    unparse_rhs(X) ++ ", " ++ unparse_rhs(Xs);
unparse_rhs(Nt) when is_atom(Nt) ->
    atom_to_list(Nt);
unparse_rhs(#pattern{op=Op, kids=[]}) ->
    atom_to_list(Op);
unparse_rhs(#pattern{op=Op, kids=Kids}) ->
    atom_to_list(Op) ++ "(" ++ unparse_rhs(Kids) ++ ")".


%%%_. Emacs
%%% Local Variables:
%%% mode: erlang
%%% allout-layout: t
%%% time-stamp-pattern: "Modified : %%$"
%%% End:
