%%%============================================================================
%%% File     : eburg_util.erl
%%% Purpose  : Utility procedures.
%%% Author   : j@svirfneblin.org
%%% License  : BSD
%%% Created  : 2007-09-20 13:00:39 j
%%% Modified : 2007-11-01 18:27:07 j
%%%============================================================================

%%%_. Attributes

-module(eburg_util).
-author('j@svirfneblin.org').

-export([die/1, maybe_die/2]).
-export([say/1, say/2, psay/1, ssay/2, sayforeach/3, join/2, join/3]).
-export([butlast/1, uniq/1, enumerate/1, gensym/1, pp/1, mk_reader/2,
         assoc/2, next_pot/1, next_moe/1]).

-include("eburg.hrl").


%%%_. Error handling
%%% `die' is for non-recoverable errors. `maybe_die' takes lists of errors and
%%% warnings and does the right thing.
%%% An `error_descriptor' is a tuple which `format_error' knows how to
%%% interpret. An `error_info' is an `error_descriptor' bundled with file and
%%% possibly location information.

%% die(Descr: error_info)
%% prints an error message and exits (by throwing `error', which is
%% only caught in `eburg:file' and hence fatal).
die(Descr) ->
    err(1, Descr),
    throw(error).


%% maybe_die(Errs: [error_info], Warns: [error_info]) -> ok
%% iterates over `Errs' and `Warns' and prints appropriate
%% messages. Exits if `Errs' is non-nil.
maybe_die(Errs, Warns) ->
    foreach(fun (W) -> err(0, W) end, Warns),
    foreach(fun (E) -> err(1, E) end, Errs),
    if Errs =/= [] -> throw(error);
       true        -> ok
    end.


%% err(Level: int, Enfo: error_info) -> ok
%% prints an Emacs-friendly error message.
err(Level, Enfo) ->
    Msg0 = if Level =:= 0 -> "Warning: ";
              true        -> ""
           end,
    case Enfo of
        {File, Line, Reason} ->
            Msg1 = Msg0 ++ format_error(Reason),
            io:format("~s:~w: ~s~n", [File, Line, Msg1]);
        {File, Reason} ->
            Msg1 = Msg0 ++ format_error(Reason),
            io:format("~s: ~s~n", [File, Msg1])
    end.


%% format_error(Reason: error_descriptor) -> string
%% returns a descriptive error message for `Reason'.
format_error({file, Reason}) ->
    file:format_error(Reason);
format_error({syntax, Module, Descr}) ->
    "bad syntax: " ++ apply(Module, format_error, [Descr]);
format_error({code, Module, Descr}) ->
    "bad user code: " ++ apply(Module, format_error, [Descr]);
format_error({action, Module, Descr}) ->
    "bad semantic action: " ++ apply(Module, format_error, [Descr]);
format_error({cost, Module, Descr}) ->
    "bad dynamic cost: " ++ apply(Module, format_error, [Descr]);
format_error({dup, Ops}) ->
    io_lib:format("multiple operator declarations: ~p", [Ops]);
format_error(start) ->
    "no rules for declared start symbol";
format_error({use, Op}) ->
    io_lib:format("declared as operator but used as nonterminal: ~p", [Op]);
format_error({undeclared, Op}) ->
    io_lib:format("used as operator but not declared: ~p", [Op]);
format_error({arity, Op, Ns}) ->
    io_lib:format("operator used with different arities: ~p: ~p", [Op, Ns]);
format_error({unused, Op}) ->
    io_lib:format("operator not used: ~p", [Op]);
format_error({unreachable, Nt}) ->
    io_lib:format("nonterminal not reachable from start symbol: ~p", [Nt]).


%%%_. Strings & I/O
%%% `say' is used by `eburg_backend' to print to '$out_stream'.
%%% `sayforeach' and `join' are used to handle sequences of expressions or
%%% clauses (which typically have to be separated by ",", ";", or suchlike).
%%% `psay' (plain say) is used when the string to be printed might contain
%%%  calls to `io:format'. `ssay' gives nicer output when testing from the
%%% shell than plain `io_lib:format'.

%% say(Fmt: string) -> ok
%% say(Fmt: string, Data: [term]) -> ok
%% prints `Data' formatted according to `Fmt'.
%% uses the processs dictionary.
say(Fmt)       -> io:format(get('$out_stream'), Fmt ++ "~n", []).
say(Fmt, Data) -> io:format(get('$out_stream'), Fmt ++ "~n", Data).


%% psay(Str: string) -> ok
%% prints `Str' to the BURM file.
%% uses the process dictionary.
psay(Str) -> io:put_chars(get('$out_stream'), Str).


%% ssay(Fmt: string, Data: [term]) -> string
%% formats `Data' according to `Fmt'
ssay(Fmt, Data) -> flatten(io_lib:format(Fmt, Data)).


%% sayforeach(Say: fun, Sep: string, Xs: [term]) -> ok
%% calls `Say' on each element of `Xs' in turn, printing `Sep' after every
%% element but the last.
sayforeach(  _,   _, [])     -> ok;
sayforeach(Say,   _, [X])    -> Say(X);
sayforeach(Say, Sep, [X|Xs]) -> Say(X), say(Sep), sayforeach(Say, Sep, Xs).


%% join(Sep: string, Xs: [term]) -> string
%% join(Say: fun, Sep: string, Xs: [term]) -> string
%% returns the result of linking (the results of calling `Say' on) the elements
%% of `Xs' with `Sep'.
join(Sep, Xs)          -> join(fun(X) -> X end, Sep, Xs).
join(  _,   _, [])     -> "";
join(Say,   _, [X])    -> Say(X);
join(Say, Sep, [X|Xs]) -> Say(X) ++ Sep ++ join(Say, Sep, Xs).


%%%_. Misc

%% butlast(Xs: [term]) -> [term]
%% returns `Xs' without its last element.
butlast([_X])   -> [];
butlast([X|Xs]) -> [X|butlast(Xs)].


%% uniq(Lst: [term]) -> [term]
%% returns `Lst' with duplicate elements removed (first occurrence kept).
uniq([X|Xs]) -> [X|uniq([Y || Y <- Xs, Y =/= X])];
uniq([])     -> [].


%% enumerate(Lst: [term]) -> [{term, int}]
%% returns `Lst' with sequence numbers attached to its elements.
enumerate(Lst) -> zip(Lst, seq(1, length(Lst))).


%% gensym(Base: atom) -> atom
%% gensym(Base: string) -> string
%% implements a simple gensym-like facility for strings and atoms.
%% uses the process dictionary.
gensym(Base) when is_atom(Base) ->
    Str0 = atom_to_list(Base),
    Str1 = gensym(Str0),
    list_to_atom(Str1);
gensym(Base) when is_list(Base) ->
    Sym = [$_,$_|Base],
    case get(Sym) of
        undefined -> N = 0;
        Old       -> N = Old + 1
    end,
    put(Sym, N),
    Base ++ integer_to_list(N).


%% pp(File: name) -> ok
%% pretty-prints `File'.
pp(File) -> erl_tidy:file(File, [{backups, false}, idem]).


%% mk_reader(Lexer: fun, Parser: fun) -> fun
%% returns a function which uses `Lexer' and `Parser' (assumed to follow
%% yecc/stdlib conventions) to tokenize and parse its single argument
%% (which must be a string).
mk_reader(Lexer, Parser) ->
    fun (String) ->
            case Lexer(String) of
                {error, Reason, _Line} -> {error, Reason};
                {ok, Tokens, _Line} ->
                    case Parser(Tokens) of
                        Err={error, _Reason} -> Err;
                        {ok, Result} -> Result
                    end
            end
    end.


%% assoc(Key: term, Alist: [{term, term}]) -> term | none
%% returns the value associated with `Key' in `Alist'.
assoc(Key, Alist) ->
    case proplists:lookup(Key, Alist) of
        {Key, Val} -> Val;
        none -> none
    end.


%% next_pot(N: int) -> int
%% returns the next power of two >= `N'.
next_pot(N) -> trunc(math:log(N) / math:log(2)) + 1.


%% next_moe(N: int) -> int
%% returns the next multiple of eight >= `N'.
next_moe(N) ->
    Rest = N rem 8,
    if  Rest =:= 0 -> N;
        true       -> N + (8-Rest)
    end.


%%%_. Emacs
%%% Local Variables:
%%% mode: erlang
%%% allout-layout: t
%%% time-stamp-pattern: "Modified : %%$"
%%% End:
